/*
  File name        : queue.c
  Date of creation : 25/01/2010
  Version          : 1.0
  Copyright        : Soft'n'Design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : Queue management

  History          :
  - initial version
*/

#define _GNU_SOURCE
#include <assert.h>
#include <malloc.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "core.h"
#include "rdtsc.h"
#include "verbose.h"

#include "queue.h"

#define DEBUG_PTHREAD_LOCK(statement)                                   \
    do {                                                                \
        /* VERBOSE (coretools, 3, printf ("%s:%d 0x%x %s\n", __FILE__, __LINE__, (int)pthread_self (), #statement)); */ \
        statement;                                                      \
    } while (0);

void *queue_runner (void *param)
{

    int i, j;
    int rc;

    queue_task_t *t = (queue_task_t *)param;

    /* find queue id */
    queue_task_info_t *sq = NULL;
    int sq_id;
    for (sq_id = 0; sq_id < t->nb_sub_queues; sq_id++) {
        sq = t->sub_queue_list + sq_id;
        if ((sq) && (sq->thread == pthread_self ()))
            break;
    }
    assert (sq);

    /* initialization */
    sq->state = QUEUE_INIT;
    if (t->init)
        t->init ();

    sq->state = QUEUE_WAITING;
    sq->nb_calls = sq->nb_execs = 0;

    // choice of one CPU in t->cpu_list
    cpu_set_t cpuset;
    CPU_ZERO (&cpuset);
    char *cpu, *current;
    char *cpu_list = current = strdup (t->cpu_list);
    while ((cpu = strtok_r (NULL, "+, ", &current)) != NULL) {
        CPU_SET (atoi (cpu), &cpuset);
    }

    VERBOSE (coretools, 1, printf ("%s: %d cpus selected\n", t->name, CPU_COUNT (&cpuset)));

    int selected_cpu = -1;
    if (sq_id < CPU_COUNT (&cpuset)) {
        for (i = 0, j = 0;; i++) {
            if (CPU_ISSET (i, &cpuset)) {
                if (j == sq_id) {
                    selected_cpu = i;
                    VERBOSE (coretools, 0, printf ("%s: select cpu n-%d for sub-queue %d\n", t->name, i, j));
                    break;
                }
                j++;
            }
        }
    }
    if (selected_cpu < 0) {
        VERBOSE (coretools, 0, printf ("%s: ending sub-queue %d (%lu)\n", t->name, sq_id, pthread_self ()));
        pthread_exit (NULL);
    }
    VERBOSE (coretools, 1, printf ("%s: selected cpu %d\n", t->name, selected_cpu));

    CPU_ZERO (&cpuset);
    CPU_SET (selected_cpu, &cpuset);
    rc = pthread_setaffinity_np (pthread_self (), sizeof (cpu_set_t), &cpuset);
    if (rc) {
        VERBOSE (coretools, 0, printf ("%s: pthread_setaffinity_np error for %s\n", t->name, cpu_list));
    }
    free (cpu_list);

    struct sched_param thread_param;
    thread_param.__sched_priority = t->priority;
    rc = pthread_setschedparam (pthread_self (), SCHED_FIFO, &thread_param);
    if (rc) {
        VERBOSE (coretools, 0, printf ("%s: error enabling RT priority\n", t->name));
    }

    while (1) {
        sq->nb_calls++;

        DEBUG_PTHREAD_LOCK (pthread_mutex_lock (&t->exec_mutex));
        queue_sub_task_t *sub_task = NULL;
        for (i = 0; i < t->nb_sub_tasks; i++) {
            sub_task = t->sub_task_list + i;
            if (sub_task->state == QUEUE_LOADED)
                break;
        }
        if (i == t->nb_sub_tasks) {
            DEBUG_PTHREAD_LOCK (pthread_cond_wait (&t->wakeup_cond, &t->exec_mutex));
            DEBUG_PTHREAD_LOCK (pthread_mutex_unlock (&t->exec_mutex));
            continue;
        }
        assert (sub_task);

        sub_task->state = QUEUE_ACTIVE;
        sq->nb_execs++;
        sq->state = QUEUE_BUSY;
        DEBUG_PTHREAD_LOCK (pthread_mutex_unlock (&t->exec_mutex));

        unsigned long long tic = rdtsc ();

        if (sub_task->task)
            sub_task->task (sub_task->params);

        //__sync_add_and_fetch (t->nb_cycles + sq_id, rdtsc () - tic);
        t->nb_cycles[sq_id] += tsc2diff (rdtsc (), tic);
        __sync_add_and_fetch (t->nb_calls + sq_id, 1);

        sub_task->state = QUEUE_INACTIVE;
        sq->state = QUEUE_WAITING;

        DEBUG_PTHREAD_LOCK (pthread_rwlock_unlock (sub_task->wait_rwlock));
    }

    return NULL;
}

queue_task_t *create_queue_task (const char *name, int nb_sub_tasks, int priority,
                                 int nb_sub_queues, const char *cpu_list,
                                 void (*init) (void))
{
    if ((cpu_list == NULL) || (strlen (cpu_list) == 0))
        return NULL;

    printf ("Creating queue %s[%d/%d] prio=%d cpu=%s\n", name, nb_sub_queues, nb_sub_tasks, priority, cpu_list);

    queue_task_t *t = (queue_task_t *) calloc (1, sizeof (queue_task_t));
    assert (t);

    t->name = strdup (name);
    t->nb_sub_tasks = nb_sub_tasks;
    if (cpu_list) {
        t->cpu_list = strdup(cpu_list);
        cpu_set_t cpuset;
        CPU_ZERO (&cpuset);
        char *cpu, *current;
        char *cpu_list = current = strdup (t->cpu_list);
        while ((cpu = strtok_r (NULL, "+, ", &current)) != NULL) {
            CPU_SET (atoi (cpu), &cpuset);
        }
        free (cpu_list);
        nb_sub_queues = CPU_COUNT (&cpuset);
    }
    t->nb_sub_queues = nb_sub_queues;
    t->priority = priority;
    t->init = init;

    t->sub_queue_list =
        (queue_task_info_t *) calloc (nb_sub_queues, sizeof (queue_task_info_t));
    assert (t->sub_queue_list);

    t->sub_task_list =
        (queue_sub_task_t *) calloc (nb_sub_tasks, sizeof (queue_sub_task_t));
    assert (t->sub_task_list);

    t->nb_cycles = (unsigned long long *) calloc (nb_sub_queues, sizeof (unsigned long long));
    assert (t->nb_cycles);

    t->nb_calls = (int *) calloc (nb_sub_queues, sizeof (int));
    assert (t->nb_calls);

    t->nb_cycles_last = (unsigned long long *) calloc (nb_sub_queues, sizeof (unsigned long long));
    assert (t->nb_cycles_last);

    t->nb_calls_last = (int *) calloc (nb_sub_queues, sizeof (int));
    assert (t->nb_calls_last);

    t->nb_cycles_global = (unsigned long long *) calloc (nb_sub_queues, sizeof (unsigned long long));
    assert (t->nb_cycles_global);

    t->nb_calls_global = (int *) calloc (nb_sub_queues, sizeof (int));
    assert (t->nb_calls_global);

    /* initialize mutex and condition */
    DEBUG_PTHREAD_LOCK (pthread_mutex_init (&t->load_mutex, NULL));
    DEBUG_PTHREAD_LOCK (pthread_mutex_init (&t->exec_mutex, NULL));
    DEBUG_PTHREAD_LOCK (pthread_cond_init (&t->wakeup_cond, NULL));

    /* create threads */
    pthread_attr_t tattr;
    int i, ret = 0;
    size_t size = 1024 * 1024;
    ret |= pthread_attr_init (&tattr);
    ret |= pthread_attr_setstacksize (&tattr, size);
    for (i = 0; i < nb_sub_queues; i++) {
        pthread_create (&t->sub_queue_list[i].thread, &tattr, queue_runner, t);
    }

    return t;
}

queue_sub_task_info_t *queue_init (queue_task_t *t)
{

    queue_sub_task_info_t *st = NULL;
    if (t)
        st = (queue_sub_task_info_t *) calloc (1, sizeof (queue_sub_task_info_t));
    if (!st) return NULL;

    int i;
    int nb_sub_queues = t->nb_sub_queues;
    for (i = 0; i < nb_sub_queues; i++) {
        t->nb_cycles_global[i] += t->nb_cycles[i];
        t->nb_cycles[i] = 0;
        t->nb_calls_global[i] += t->nb_calls[i];
        t->nb_calls[i] = 0;
    }

    pthread_rwlockattr_t rwlattr;
    pthread_rwlockattr_init (&rwlattr);
    pthread_rwlockattr_setpshared (&rwlattr, PTHREAD_PROCESS_SHARED);
    DEBUG_PTHREAD_LOCK (pthread_rwlock_init (&st->wait_rwlock, &rwlattr));

    return st;
}

int queue_load (queue_task_t *t, queue_sub_task_info_t *st,
                int (*task) (void *), void *params)
{
    int ret = 0;

    if (t && st) {
        queue_sub_task_t *sub_task = NULL;
        int i;

        DEBUG_PTHREAD_LOCK (pthread_mutex_lock (&t->load_mutex));
        for (i = 0; i < t->nb_sub_tasks; i++) {
            sub_task = t->sub_task_list + i;
            if (sub_task->state == QUEUE_INACTIVE)
                break;
        }
        if (i == t->nb_sub_tasks) {
            VERBOSE (coretools, 1, printf ("queue %s is undersized\n", t->name));
            DEBUG_PTHREAD_LOCK (pthread_mutex_unlock (&t->load_mutex));
            return 0;
        }
        assert (sub_task);

        sub_task->state = QUEUE_LOADING;
        DEBUG_PTHREAD_LOCK (pthread_mutex_unlock (&t->load_mutex));

        sub_task->task = task;
        sub_task->params = params;

        sub_task->wait_rwlock = &st->wait_rwlock;
        DEBUG_PTHREAD_LOCK (pthread_rwlock_rdlock (&st->wait_rwlock));

        sub_task->state = QUEUE_LOADED;

        ret = 1;
    }

    return ret;
}

int queue_end (queue_task_t *t, queue_sub_task_info_t *st)
{

    if (st) {

        DEBUG_PTHREAD_LOCK (pthread_cond_broadcast (&t->wakeup_cond));

        DEBUG_PTHREAD_LOCK (pthread_rwlock_wrlock (&st->wait_rwlock));

        //DEBUG_PTHREAD_LOCK (pthread_rwlock_unlock (&st->wait_rwlock));

        DEBUG_PTHREAD_LOCK (pthread_rwlock_destroy (&st->wait_rwlock));

        free (st);

    }

    int i;
    int nb_sub_queues = t->nb_sub_queues;
    for (i = 0; i < nb_sub_queues; i++) {
        t->nb_cycles_last[i] = t->nb_cycles[i];
        t->nb_calls_last[i] = t->nb_calls[i];
    }

    return 1;
}

void queue_stat (queue_task_t *t, int condensed)
{

    int i;

    if (t->nb_sub_queues <= 0)
        return;

    int nb_sub_queues = t->nb_sub_queues;

    if (condensed) {
        printf ("%s=[", t->name);
        for (i = 0; i < nb_sub_queues; i++) {
            if (i != 0) printf ("|");
            printf ("%.1f(%d)", tsc2ms (t->nb_cycles_last[i]), t->nb_calls_last[i]);
        }
        printf ("]");
    } else {
        printf ("Queue: %s\n[Id] time(ms) (# cycles) / # calls\n",  t->name);
        unsigned long long nb_cycles_global = 0;
        int nb_calls_global = 0;
        for (i = 0; i < t->nb_sub_queues; i++) {
            printf ("[%d] %.1f (%llu) / %d\n", i,
                    tsc2ms (t->nb_cycles_global[i] + t->nb_cycles[i]),
                    t->nb_cycles_global[i] + t->nb_cycles[i],
                    t->nb_calls_global[i] + t->nb_calls[i]);
            nb_cycles_global += t->nb_cycles_global[i] + t->nb_cycles[i];
            nb_calls_global += t->nb_calls_global[i] + t->nb_calls[i];
        }
        unsigned long long avg_nb_cycles = nb_cycles_global / nb_calls_global;
        printf ("Average time(ms) (# cycles): %.1f (%llu)\n",
                tsc2ms (avg_nb_cycles), avg_nb_cycles);
    }

}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
